import os
import sys
import cv2
import numpy as np
import glob
from PIL import Image
import pickle


def load_npy_dict_file(fname):
    if fname is None:
        return None

    if os.path.exists(fname):
        npy_dict = np.load(fname, allow_pickle=True).item()
        assert type(npy_dict) == dict or npy_dict is None
        return npy_dict
    else:
        return None


def load_pickle_file(fname):
    if fname is None:
        return None

    if os.path.exists(fname):
        with open(fname, 'rb') as f:
            model = pickle.load(f)

        return model
    else:
        return None


# def read_image_to_np(img_path: str) -> np.ndarray:
#     img = cv2.imread(img_path, cv2.IMREAD_COLOR)
#     img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)  # convert BGR to RGB.
#     # h, w, c = img.shape
#     # assert c in [1, 3]
#     # assert c.dytpe == np.unit8
#     return img


def resize_image(img: np.ndarray, target_size: int, interpolation=cv2.INTER_LINEAR) -> np.ndarray:
    """resize image with shorter edge equal to target_size.

    Args:
        img: image data
        target_size: resize short target size
        interpolation: interpolation mode

    Returns:
        resized image data
    """

    h, w, c = img.shape
    assert c in [1, 3]

    percent = float(target_size) / min(h, w)
    resized_width = int(round(w * percent))
    resized_height = int(round(h * percent))

    resized = cv2.resize(img, (resized_width, resized_height), interpolation=interpolation)
    # assert resized.dytpe == np.float32
    return resized


def crop_image(img: np.ndarray, target_size: int, center=True) -> np.ndarray:
    """crop image

    Args:
        img: images data
        target_size: crop target size
        center: crop mode

    Returns:
        img: cropped image data
    """
    height, width = img.shape[:2]
    size = target_size
    if center:
        w_start = (width - size) // 2
        h_start = (height - size) // 2
    else:
        w_start = np.random.randint(0, width - size + 1)
        h_start = np.random.randint(0, height - size + 1)
    w_end = w_start + size
    h_end = h_start + size
    img = img[h_start:h_end, w_start:w_end, :]
    return img


def preprocess_image(img: np.ndarray, random_mirror=False) -> np.ndarray:
    """
    image(uint8) to tensor(float32). scaled by 1/255, centered, standarized.
    :param img: np.ndarray: shape: [ns, h, w, 3], color order: rgb.
    :return: np.ndarray: shape: [ns, 3, h, w]
    """
    # ImageNet stats.
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # transpose from [ns, h, w, 3] to [ns, c, h, w], and scaled by 1/255
    img = img.astype('float32').transpose((0, 3, 1, 2)) / 255

    img_mean = np.array(mean).reshape((3, 1, 1))
    img_std = np.array(std).reshape((3, 1, 1))
    img -= img_mean
    img /= img_std

    if random_mirror:
        mirror = int(np.random.uniform(0, 2))
        if mirror == 1:
            img = img[:, :, ::-1, :]

    return img


def read_image(img_path, target_size=256, crop_size=224, crop=True) -> np.ndarray:
    """
    resize_short to target_size, then center crop to crop_size or not crop.
    :param img_path: one image path
    :return: np.ndarray: shape: [1, h, w, 3], dtype: uint8, color order: rgb.
    """

    if isinstance(img_path, str):
        with open(img_path, 'rb') as f:
            img = Image.open(f)
            img = img.convert('RGB')
            img = np.array(img)
            img = resize_image(img, target_size)
            if crop:
                img = crop_image(img, target_size=crop_size, center=True)
            img = np.expand_dims(img, axis=0)
            return img
    elif isinstance(img_path, np.ndarray):
        assert len(img_path.shape) == 4
        return img_path
    else:
        ValueError(f"Not recognized data type {type(img_path)}.")


def restore_image(float_input_data: np.ndarray) -> np.ndarray:
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]
    img_mean = np.array(mean).reshape((3, 1, 1))
    img_std = np.array(std).reshape((3, 1, 1))
    float_input_data *= img_std
    float_input_data += img_mean
    float_input_data *= 255
    float_input_data += 0.5  # for float to integer
    img = np.uint8(float_input_data.transpose((0, 2, 3, 1)))
    return img


# def _find_classes(dir):
#     # Faster and available in Python 3.5 and above
#     classes = [d.name for d in os.scandir(dir) if d.is_dir()]
#     classes.sort()
#     class_to_idx = {classes[i]: i for i in range(len(classes))}
#     return classes, class_to_idx


# def get_typical_dataset_info(dataset_dir,
#                              subset="test",
#                              shuffle=False,
#                              random_seed=None):
#     """
#     where {dataset_dir}/{train,test,segmentations}}/{class1, class2, ...}/*.png exists.

#     segmentations are optional.

#     Args:
#         dataset_dir:
#         shuffle:
#         random_seed:

#     Returns:

#     """
#     IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif',
#                       '.tiff', '.webp')

#     # read
#     set_dir = os.path.join(dataset_dir, subset)
#     seg_dir = os.path.join(dataset_dir, 'segmentations')

#     class_names, class_to_idx = _find_classes(set_dir)
#     # num_classes = len(class_names)
#     image_paths = []
#     seg_paths = []
#     labels = []
#     for class_name in sorted(class_names):
#         classes_dir = os.path.join(set_dir, class_name)
#         for img_path in sorted(glob.glob(os.path.join(classes_dir, '*'))):
#             if not img_path.lower().endswith(IMG_EXTENSIONS):
#                 continue

#             image_paths.append(img_path)
#             seg_paths.append(
#                 os.path.join(seg_dir,
#                              img_path.split('test/')[-1].replace('jpg',
#                                                                  'png')))
#             labels.append(class_to_idx[class_name])

#             assert os.path.exists(seg_paths[-1]), seg_paths[-1]

#     image_paths = np.array(image_paths)
#     seg_paths = np.array(seg_paths)
#     labels = np.array(labels)

#     if shuffle:
#         np.random.seed(random_seed)
#         random_per = np.random.permutation(range(len(image_paths)))
#         image_paths = image_paths[random_per]
#         seg_paths = seg_paths[random_per]
#         labels = labels[random_per]

#     return image_paths, seg_paths, labels, len(class_names)


# def extract_img_paths(directory):
#     IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif',
#                       '.tiff', '.webp')
#     img_paths = []
#     img_names = []
#     for file in os.listdir(directory):
#         if file.lower().endswith(IMG_EXTENSIONS):
#             img_paths.append(os.path.join(directory, file))
#             img_names.append(file)
#     return img_paths, img_names



def images_transform_pipeline(array_or_path, resize_to=224, crop_to=None):
    """[summary]

    Args:
        array_or_path ([type]): [description]
        resize_to (int, optional): [description]. Defaults to 224.
        crop_to ([type], optional): [description]. Defaults to None.
    """
    if crop_to is not None:
        assert isinstance(crop_to, int)
        def read_image_func(path):
            return read_image(path, target_size=resize_to, crop_size=crop_to)
    else:
        def read_image_func(path):
            return read_image(path, target_size=resize_to, crop=False)

    if isinstance(array_or_path, str):
        # one single image path.
        uint8_imgs = read_image_func(array_or_path)
        float_input_data = preprocess_image(uint8_imgs)
    elif isinstance(array_or_path, list) and all(
            isinstance(elem, str) for elem in array_or_path):
        # a list of image paths.
        uint8_imgs = []
        for fp in array_or_path:
            uint8_img = read_image_func(fp)
            uint8_imgs.append(uint8_img)
        uint8_imgs = np.concatenate(uint8_imgs)
        float_input_data = preprocess_image(uint8_imgs)
    else:
        # an array. will not resize nor crop.
        if len(array_or_path.shape) == 3:
            uint8_imgs = np.expand_dims(array_or_path, axis=0)
        elif len(array_or_path.shape) == 4:
            uint8_imgs = array_or_path

        if np.issubdtype(array_or_path.dtype, np.integer):
            # array_or_path is an image.
            uint8_imgs = uint8_imgs.copy()
            float_input_data = preprocess_image(uint8_imgs)
        else:
            # array_or_path is float input data.
            uint8_imgs = restore_image(array_or_path.copy())
            float_input_data = array_or_path

    return uint8_imgs, float_input_data


# def preprocess_inputs(inputs, model_input_shape):
#     """[summary]

#     Args:
#         inputs ([type]): can be a str, a list of str, or np.ndarray.
#         model_input_shape ([type]): [description]

#     Returns:
#         imgs: uint8 images, used for visualization. [ns, h, w, 3]
#         data: float scaled image data, used for computation. [ns, 3, h, w]
#     """
#     if isinstance(inputs, str):
#         imgs = read_image(inputs, crop_size=model_input_shape[1])
#         data = preprocess_image(imgs)
#     elif isinstance(inputs, list) and all(
#             isinstance(elem, str) for elem in inputs):
#         imgs = []
#         for fp in inputs:
#             img = read_image(fp, crop_size=model_input_shape[1])
#             imgs.append(img)
#         imgs = np.concatenate(imgs)
#         data = preprocess_image(imgs)
#     else:
#         if len(inputs.shape) == 3:
#             inputs = np.expand_dims(inputs, axis=0)
#         if np.issubdtype(inputs.dtype, np.integer):
#             imgs = inputs.copy()
#             data = preprocess_image(inputs)
#         else:
#             imgs = restore_image(inputs.copy())
#             data = inputs
#     return imgs, data


def preprocess_save_path(save_path, bsz):
    if isinstance(save_path, str):
        save_path = [save_path]
    if save_path is None:
        save_path = [None] * bsz
    assert len(
        save_path
    ) == bsz, f"number of save_paths ({len(save_path)}) should be equal to number of images ({bsz})"
    return save_path



from copy import Error
import paddle.nn as nn
import paddle

from ._layers_lrp import *


__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152']


model_urls = {
    'resnet18': ('https://paddle-hapi.bj.bcebos.com/models/resnet18.pdparams',
                 'cf548f46534aa3560945be4b95cd11c4'),
    'resnet34': ('https://paddle-hapi.bj.bcebos.com/models/resnet34.pdparams',
                 '8d2275cf8706028345f78ac0e1d31969'),
    'resnet50': ('https://paddle-hapi.bj.bcebos.com/models/resnet50.pdparams',
                 'ca6f485ee1ab0492d38f323885b0ad80'),
    'resnet101': ('https://paddle-hapi.bj.bcebos.com/models/resnet101.pdparams',
                  '02f35f034ca3858e1e54d4036443c92d'),
    'resnet152': ('https://paddle-hapi.bj.bcebos.com/models/resnet152.pdparams',
                  '7ad16a2f1e7333859ff986138630fd7a'),
}


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return Conv2D(in_planes, out_planes, kernel_size=3, stride=stride,
                  padding=1, bias_attr=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return Conv2D(in_planes, out_planes, kernel_size=1, stride=stride, bias_attr=False)


class BasicBlock(nn.Layer):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.clone = Clone()

        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = BatchNorm2D(planes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = BatchNorm2D(planes)
        self.downsample = downsample
        self.stride = stride

        self.relu1 = ReLU()
        self.relu2 = ReLU()

        self.add = Add()

        self.register_forward_post_hook(forward_hook)

    def forward(self, x):
        x1, x2 = self.clone(x, 2)

        out = self.conv1(x1)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            x2 = self.downsample(x2)

        out = self.add([out, x2])
        out = self.relu2(out)

        return out

    def relprop(self, R, alpha):
        out = self.relu2.relprop(R, alpha)
        out, x2 = self.add.relprop(out, alpha)

        if self.downsample is not None:
            x2 = self.downsample.relprop(x2, alpha)

        out = self.bn2.relprop(out, alpha)
        out = self.conv2.relprop(out, alpha)

        out = self.relu1.relprop(out, alpha)
        out = self.bn1.relprop(out, alpha)
        x1 = self.conv1.relprop(out, alpha)

        return self.clone.relprop([x1, x2], alpha)

    def m_relprop(self, R, pred, alpha):
        out = self.relu2.m_relprop(R, pred, alpha)
        out, x2 = self.add.m_relprop(out, pred, alpha)

        if self.downsample is not None:
            x2 = self.downsample.m_relprop(x2, pred, alpha)

        out = self.bn2.m_relprop(out, pred, alpha)
        out = self.conv2.m_relprop(out, pred, alpha)

        out = self.relu1.m_relprop(out, pred, alpha)
        out = self.bn1.m_relprop(out, pred, alpha)
        x1 = self.conv1.m_relprop(out, pred, alpha)

        return self.clone.m_relprop([x1, x2], pred, alpha)

    def RAP_relprop(self, R):
        out = self.relu2.RAP_relprop(R)
        out, x2 = self.add.RAP_relprop(out)

        if self.downsample is not None:
            x2 = self.downsample.RAP_relprop(x2)

        out = self.bn2.RAP_relprop(out)
        out = self.conv2.RAP_relprop(out)

        out = self.relu1.RAP_relprop(out)
        out = self.bn1.RAP_relprop(out)
        x1 = self.conv1.RAP_relprop(out)

        return self.clone.RAP_relprop([x1, x2])


class Bottleneck(nn.Layer):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.clone = Clone()

        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = BatchNorm2D(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = BatchNorm2D(planes)
        self.conv3 = conv1x1(planes, planes * self.expansion)
        self.bn3 = BatchNorm2D(planes * self.expansion)
        self.downsample = downsample
        self.stride = stride

        self.relu1 = ReLU()
        self.relu2 = ReLU()
        self.relu3 = ReLU()

        self.add = Add()

        self.register_forward_post_hook(forward_hook)

    def forward(self, x):
        # x1, x2 = self.clone(x, 2)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            x = self.downsample(x)

        # out = self.add([out, x2])
        out = self.add([out, x])
        out = self.relu3(out)

        return out

    def relprop(self, R, alpha):
        out = self.relu3.relprop(R, alpha)

        out, x = self.add.relprop(out, alpha)

        if self.downsample is not None:
            x = self.downsample.relprop(x, alpha)

        out = self.bn3.relprop(out, alpha)
        out = self.conv3.relprop(out, alpha)

        out = self.relu2.relprop(out, alpha)
        out = self.bn2.relprop(out, alpha)
        out = self.conv2.relprop(out, alpha)

        out = self.relu1.relprop(out, alpha)
        out = self.bn1.relprop(out, alpha)
        x1 = self.conv1.relprop(out, alpha)

        return x1 + x
        # return self.clone.relprop([x1, x2], alpha)

    def m_relprop(self, R, pred, alpha):
        out = self.relu3.m_relprop(R, pred, alpha)

        out, x = self.add.m_relprop(out, pred, alpha)

        if self.downsample is not None:
            x = self.downsample.m_relprop(x, pred, alpha)

        out = self.bn3.m_relprop(out, pred, alpha)
        out = self.conv3.m_relprop(out, pred, alpha)

        out = self.relu2.m_relprop(out, pred, alpha)
        out = self.bn2.m_relprop(out, pred, alpha)
        out = self.conv2.m_relprop(out, pred, alpha)

        out = self.relu1.m_relprop(out, pred, alpha)
        out = self.bn1.m_relprop(out, pred, alpha)
        x1 = self.conv1.m_relprop(out, pred, alpha)
        if paddle.is_tensor(x1) == True:
            return x1 + x
        else:
            for i in range(len(x1)):
                x1[i] = x1[i] + x[i]
            return x1

    def RAP_relprop(self, R):
        out = self.relu3.RAP_relprop(R)

        out, x = self.add.RAP_relprop(out)

        if self.downsample is not None:
            x = self.downsample.RAP_relprop(x)

        out = self.bn3.RAP_relprop(out)
        out = self.conv3.RAP_relprop(out)

        out = self.relu2.RAP_relprop(out)
        out = self.bn2.RAP_relprop(out)
        out = self.conv2.RAP_relprop(out)

        out = self.relu1.RAP_relprop(out)
        out = self.bn1.RAP_relprop(out)
        x1 = self.conv1.RAP_relprop(out)

        return x1 + x


class ResNet(nn.Layer):

    def __init__(self, block, layers, num_classes=1000, zero_init_residual=False):
        super(ResNet, self).__init__()
        self.inplanes = 64
        self.conv1 = Conv2D(3, 64, kernel_size=7, stride=2, padding=3,
                            bias_attr=False)
        self.bn1 = BatchNorm2D(64)
        self.relu = ReLU()
        self.maxpool = MaxPool2D(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = AdaptiveAvgPool2D((1, 1))
        self.fc = Linear(512 * block.expansion, num_classes)

        # for m in [self] + self.sublayers():
        #     if isinstance(m, nn.Conv2D):
        #         nn.init.kaiming_normal_(
        #             m.weight, mode='fan_out', nonlinearity='relu')
        #     elif isinstance(m, nn.BatchNorm2D):
        #         nn.init.constant_(m.weight, 1)
        #         nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in [self] + self.sublayers():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                BatchNorm2D(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.reshape((x.shape[0], -1))
        x = self.fc(x)

        return x

    def relprop(self, R, alpha):
        R = self.fc.relprop(R, alpha)
        R = R.reshape(self.avgpool.Y.shape)
        R = self.avgpool.relprop(R, alpha)

        R = self.layer4.relprop(R, alpha)
        R = self.layer3.relprop(R, alpha)
        R = self.layer2.relprop(R, alpha)
        R = self.layer1.relprop(R, alpha)

        R = self.maxpool.relprop(R, alpha)
        R = self.relu.relprop(R, alpha)
        R = self.bn1.relprop(R, alpha)
        R = self.conv1.relprop(R, alpha)

        return R

    def m_relprop(self, R, pred, alpha):
        R = self.fc.m_relprop(R, pred, alpha)
        if paddle.is_tensor(R) == False:
            for i in range(len(R)):
                R[i] = R[i].reshape_as(self.avgpool.Y)
        else:
            R = R.reshape_as(self.avgpool.Y)
        R = self.avgpool.m_relprop(R, pred, alpha)

        R = self.layer4.m_relprop(R, pred, alpha)
        R = self.layer3.m_relprop(R, pred, alpha)
        R = self.layer2.m_relprop(R, pred, alpha)
        R = self.layer1.m_relprop(R, pred, alpha)

        R = self.maxpool.m_relprop(R, pred, alpha)
        R = self.relu.m_relprop(R, pred, alpha)
        R = self.bn1.m_relprop(R, pred, alpha)
        R = self.conv1.m_relprop(R, pred, alpha)

        return R

    def RAP_relprop(self, R):
        R = self.fc.RAP_relprop(R)
        R = R.reshape_as(self.avgpool.Y)
        R = self.avgpool.RAP_relprop(R)

        R = self.layer4.RAP_relprop(R)
        R = self.layer3.RAP_relprop(R)
        R = self.layer2.RAP_relprop(R)
        R = self.layer1.RAP_relprop(R)

        R = self.maxpool.RAP_relprop(R)
        R = self.relu.RAP_relprop(R)
        R = self.bn1.RAP_relprop(R)
        R = self.conv1.RAP_relprop(R)

        return R


def resnet18(pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    raise NameError("TODO: fix bugs")
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['resnet18'][0],
            model_urls['resnet18'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model


def resnet34(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    raise NameError("TODO: fix bugs")
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['resnet34'][0],
            model_urls['resnet34'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['resnet50'][0],
            model_urls['resnet50'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model


def resnet101(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['resnet101'][0],
            model_urls['resnet101'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model


def resnet152(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['resnet152'][0],
            model_urls['resnet152'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model



import abc
import sys
import numpy as np
import warnings


# Ensure compatibility with Python 2/3
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class InterpreterEvaluator(ABC):
    """Base class for interpreter evaluators.

    Args:
        ABC ([type]): [description]
    """
    def __init__(self, paddle_model: callable, device: str, use_cuda: bool, **kwargs):
        
        if use_cuda in [True, False]:
            warnings.warn(
                '``use_cuda`` would be deprecated soon. '
                'Use ``device`` directly.',
                stacklevel=2
            )
            if use_cuda and device[:3] == 'gpu':
                device = device
            else:
                device = 'cpu'
        self.device = device
        self.paddle_model = paddle_model
        self.predict_fn = None
    
    def evaluate(self, **kwargs):
        raise NotImplementedError


import paddle.nn as nn
import paddle

from ._layers_lrp import *

__all__ = [
    'VGG', 'vgg16'
]


model_urls = {
    'vgg16': ('https://paddle-hapi.bj.bcebos.com/models/vgg16.pdparams',
              '89bbffc0f87d260be9b8cdc169c991c4')
}


class VGG(nn.Layer):

    def __init__(self, features, num_classes=1000):
        super(VGG, self).__init__()
        self.features = features
        self.avgpool = AdaptiveAvgPool2D((7, 7))
        self.classifier = Sequential(
            Linear(512 * 7 * 7, 4096),
            ReLU(True),
            Dropout(),
            Linear(4096, 4096),
            ReLU(True),
            Dropout(),
            Linear(4096, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = self.avgpool(x)
        x = x.reshape((x.shape[0], -1))
        x = self.classifier(x)
        return x

    def relprop(self, R, alpha):
        x = self.classifier.relprop(R, alpha)
        x = x.reshape(next(reversed(self.features._sub_layers.values())).Y.shape)
        x = self.avgpool.relprop(x, alpha)
        x = self.features.relprop(x, alpha)

        return x
    def m_relprop(self, R, pred, alpha):
        x = self.classifier.m_relprop(R, pred, alpha)
        if paddle.is_tensor(x) == False:
            for i in range(len(x)):
                x[i] = x[i].reshape(next(reversed(self.features._sub_layers.values())).Y.shape)
        else:
            x = x.reshape(next(reversed(self.features._sub_layers.values())).Y.shape)
        x = self.avgpool.m_relprop(x, pred, alpha)
        x = self.features.m_relprop(x, pred, alpha)

        return x
    def RAP_relprop(self, R):
        x1 = self.classifier.RAP_relprop(R)
        if paddle.is_tensor(x1) == False:
            for i in range(len(x1)):
                x1[i] = x1[i].reshape(next(reversed(self.features._sub_layers.values())).Y.shape)
        else:
            x1 = x1.reshape(next(reversed(self.features._sub_layers.values())).Y.shape)
        x1 = self.avgpool.RAP_relprop(x1)
        x1 = self.features.RAP_relprop(x1)

        return x1


def make_layers(cfg, batch_norm=False):
    layers = []
    in_channels = 3
    for v in cfg:
        if v == 'M':
            layers += [MaxPool2D(kernel_size=2, stride=2)]
        else:
            conv2d = Conv2D(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, BatchNorm2D(v), ReLU()]
            else:
                layers += [conv2d, ReLU()]
            in_channels = v
    return Sequential(*layers)


cfg = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


def vgg16(pretrained=False, **kwargs):
    """VGG 16-layer model (configuration "D")

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = VGG(make_layers(cfg['D']), **kwargs)
    if pretrained:
        weight_path = paddle.utils.download.get_weights_path_from_url(
            model_urls['vgg16'][0],
            model_urls['vgg16'][1]
        )
        param = paddle.load(weight_path)
        model.set_dict(param)
    return model




# Files within this path, contain the interpretation algorithms.

from .abc_interpreter import Interpreter
from .abc_interpreter import InputGradientInterpreter, InputOutputInterpreter, IntermediateLayerInterpreter

from .lime import LIMECVInterpreter, LIMENLPInterpreter
from .gradient_cam import GradCAMInterpreter
from .integrated_gradients import IntGradCVInterpreter, IntGradNLPInterpreter
from .smooth_grad import SmoothGradInterpreter
from .smooth_grad_v2 import SmoothGradInterpreterV2
from .occlusion import OcclusionInterpreter
from .gradient_shap import GradShapCVInterpreter, GradShapNLPInterpreter
from ._normlime_base import NormLIMECVInterpreter, NormLIMENLPInterpreter
from .score_cam import ScoreCAMInterpreter
from .lrp import LRPCVInterpreter
from .rollout import RolloutInterpreter
from .transition_attention_maps import TAMInterpreter
from .consensus import ConsensusInterpreter

__all__ = [
    "Interpreter",
    "InputGradientInterpreter", "InputOutputInterpreter", "IntermediateLayerInterpreter",
    "LIMECVInterpreter", "LIMENLPInterpreter", "NormLIMECVInterpreter",
    "NormLIMENLPInterpreter", "GradCAMInterpreter", "IntGradCVInterpreter",
    "IntGradNLPInterpreter", "SmoothGradInterpreter", "OcclusionInterpreter",
    "GradShapCVInterpreter", "GradShapNLPInterpreter", "ScoreCAMInterpreter",
    "LRPCVInterpreter", "RolloutInterpreter", "TAMInterpreter",
    "SmoothGradInterpreterV2", 'ConsensusInterpreter'
]

try:
    import paddle
    from . import lime_prior
    from .lime_prior import LIMEPriorInterpreter
    from .forgetting_events import ForgettingEventsInterpreter
    __all__.append("LIMEPriorInterpreter")
    __all__.append("ForgettingEventsInterpreter")
except ModuleNotFoundError:
    print(
        "Warning: Paddle should be installed before using LIMEPriorInterpreter or ForgettingEventsInterpreter."
    )



import unittest
from paddle.vision.models import mobilenet_v2, resnet50
import numpy as np
import os
import interpretdl as it
from tests.utils import assert_arrays_almost_equal


class TestScoreCAM(unittest.TestCase):
    
    def test_shape(self):
        paddle_model = mobilenet_v2(pretrained=True)

        img_path = 'imgs/catdog.jpg'
        algo = it.ScoreCAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, 'features.18.2', resize_to=64, crop_to=64, visual=False)

        result = np.array([*exp.shape])
        desired = np.array([1.        , 64.        , 64.        ])

        assert_arrays_almost_equal(self, result, desired)

    def test_algo(self):
        paddle_model = mobilenet_v2(pretrained=True)

        np.random.seed(42)
        img_path = np.random.randint(0, 255, size=(1, 64, 64, 3), dtype=np.uint8)
        algo = it.ScoreCAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, 'features.16.conv.3', resize_to=64, crop_to=64, visual=False)

        result = np.array([exp.mean(), exp.std(), exp.min(), exp.max()])
        desired = np.array([ 2.3779001 ,  1.80234987, -0.33591489,  5.09665543])

        assert_arrays_almost_equal(self, result, desired)

    def test_save(self):
        import matplotlib
        matplotlib.use('agg')  # non-GUI, for skipping.

        paddle_model = mobilenet_v2(pretrained=True)

        np.random.seed(42)
        img_path = np.random.randint(0, 255, size=(1, 64, 64, 3), dtype=np.uint8)
        algo = it.ScoreCAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, 'features.16.conv.3', resize_to=64, crop_to=64, 
            visual=True, save_path='tmp.jpg')

        result = np.array([exp.mean(), exp.std(), exp.min(), exp.max()])
        desired = np.array([ 2.3779001 ,  1.80234987, -0.33591489,  5.09665543])

        assert_arrays_almost_equal(self, result, desired)
        os.remove('tmp.jpg')

if __name__ == '__main__':
    unittest.main()


from paddle.fluid.layers.nn import crop
from tqdm import tqdm
import numpy as np
import paddle

from .abc_interpreter import Interpreter, InputOutputInterpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image



class OcclusionInterpreter(InputOutputInterpreter):
    """
    Occlusion Interpreter.

    More details regarding the Occlusion method can be found in the original paper:
    https://arxiv.org/abs/1311.2901

    Part of the code is modified from https://github.com/pytorch/captum/blob/master/captum/attr/_core/occlusion.py
    """

    def __init__(self,
                 paddle_model,
                 use_cuda=None,
                 device='gpu:0') -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        InputOutputInterpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(self,
                  inputs,
                  sliding_window_shapes,
                  labels=None,
                  strides=1,
                  baselines=None,
                  perturbations_per_eval=1,
                  resize_to=224, 
                  crop_to=None,
                  visual=True,
                  save_path=None):
        """
        Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            sliding_window_shapes (tuple): Shape of sliding windows to occlude data.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. If None, the most likely label for each image will be used. Default: None            baseline (numpy.ndarray, optional): The baseline input. If None, all zeros will be used. Default: None
            strides (int or tuple): The step by which the occlusion should be shifted by in each direction for each iteration.
                                    If int, the step size in each direction will be the same. Default: 1
            baselines (numpy.ndarray or None, optional): The baseline images to compare with. It should have the same shape as images.
                                                        If None, the baselines of all zeros will be used. Default: None.
            perturbations_per_eval (int, optional): number of occlusions in each batch. Default: 1
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        Returns:
            [numpy.ndarray]: interpretations for images
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)

        bsz = len(data)
        save_path = preprocess_save_path(save_path, bsz)

        self._build_predict_fn(output='probability')

        if baselines is None:
            baselines = np.zeros_like(data)
        elif np.array(baselines).ndim == 3:
            baselines = np.repeat(np.expand_dims(baselines, 0), len(data), 0)
        if len(baselines) == 1:
            baselines = np.repeat(baselines, len(data), 0)

        probs, _ = self.predict_fn(data, None)

        sliding_windows = np.ones(sliding_window_shapes)

        if labels is None:
            labels = np.argmax(probs, axis=1)
        elif isinstance(labels, int):
            labels = [labels]

        img_size = [3, crop_to, crop_to] if crop_to is not None else [3, imgs.shape[1], imgs.shape[2]]
        current_shape = np.subtract(
            img_size, sliding_window_shapes
        )
        shift_counts = tuple(
            np.add(np.ceil(np.divide(current_shape, strides)).astype(int), 1)
        )

        initial_eval = np.array(
            [probs[i][labels[i]] for i in range(bsz)]).reshape((1, bsz))
        total_interp = np.zeros_like(data)

        num_features = np.prod(shift_counts)
        with tqdm(total=num_features, leave=True, position=0) as pbar:
            for (ablated_features, current_mask) in self._ablation_generator(
                    data, sliding_windows, strides, baselines, shift_counts,
                    perturbations_per_eval):
                ablated_features = ablated_features.reshape(
                    ( -1, ) + ablated_features.shape[2:]
                )
                modified_probs, _ = self.predict_fn(np.float32(ablated_features), None)
                modified_eval = [
                    p[labels[i % bsz]] for i, p in enumerate(modified_probs)
                ]
                eval_diff = initial_eval - np.array(modified_eval).reshape(
                    (-1, bsz)
                )
                eval_diff = eval_diff.T
                dim_tuple = (len(current_mask), ) + (1, ) * (current_mask.ndim - 1)
                for i, diffs in enumerate(eval_diff):
                    #j = i % perturbations_per_eval
                    total_interp[i] += np.sum(
                        diffs.reshape(dim_tuple) * current_mask, axis=0
                    )[0]
                
                pbar.update(1)

        # visualization and save image.
        for i in range(len(data)):
            vis_explanation = explanation_to_vis(imgs[i], np.abs(total_interp[i]).sum(0), style='overlay_grayscale')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        return total_interp

    def _ablation_generator(self, inputs, sliding_window, strides, baselines,
                            shift_counts, perturbations_per_eval):
        num_features = np.prod(shift_counts)
        perturbations_per_eval = min(perturbations_per_eval, num_features)
        num_features_processed = 0
        num_examples = len(inputs)
        if perturbations_per_eval > 1:
            all_features_repeated = np.repeat(
                np.expand_dims(inputs, 0), perturbations_per_eval, axis=0)
        else:
            all_features_repeated = np.expand_dims(inputs, 0)

        while num_features_processed < num_features:
            current_num_ablated_features = min(
                perturbations_per_eval, num_features - num_features_processed)
            if current_num_ablated_features != perturbations_per_eval:
                current_features = all_features_repeated[:
                                                         current_num_ablated_features]
            else:
                current_features = all_features_repeated

            ablated_features, current_mask = self._construct_ablated_input(
                current_features, baselines, num_features_processed,
                num_features_processed + current_num_ablated_features,
                sliding_window, strides, shift_counts)

            yield ablated_features, current_mask
            num_features_processed += current_num_ablated_features

    def _construct_ablated_input(self, inputs, baselines, start_feature,
                                 end_feature, sliding_window, strides,
                                 shift_counts):
        input_masks = np.array([
            self._occlusion_mask(inputs, j, sliding_window, strides,
                                 shift_counts)
            for j in range(start_feature, end_feature)
        ])
        ablated_tensor = inputs * (1 - input_masks) + baselines * input_masks

        return ablated_tensor, input_masks

    def _occlusion_mask(self, inputs, ablated_feature_num, sliding_window,
                        strides, shift_counts):
        remaining_total = ablated_feature_num

        current_index = []
        for i, shift_count in enumerate(shift_counts):
            stride = strides[i] if isinstance(strides, tuple) else strides
            current_index.append((remaining_total % shift_count) * stride)
            remaining_total = remaining_total // shift_count

        remaining_padding = np.subtract(
            inputs.shape[2:], np.add(current_index, sliding_window.shape))

        slicers = []
        for i, p in enumerate(remaining_padding):
            # When there is no enough space for sliding window, truncate the window
            if p < 0:
                slicer = [slice(None)] * len(sliding_window.shape)
                slicer[i] = range(sliding_window.shape[i] + p)
                slicers.append(slicer)

        pad_values = tuple(
            tuple(reversed(np.maximum(pair, 0)))
            for pair in zip(remaining_padding, current_index))

        for slicer in slicers:
            sliding_window = sliding_window[tuple(slicer)]
        padded = np.pad(sliding_window, pad_values)

        return padded.reshape((1, ) + padded.shape)




import numpy as np
import paddle
import re

from .abc_interpreter import Interpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image


class TAMInterpreter(Interpreter):
    """
    Transition Attention Maps Interpreter.

    More details regarding the Transition_Attention_Maps method can be found in the original paper:
    
    .. code-block:: none

        Tingyi Yuan, Xuhong Li, Haoyi Xiong, Hui Cao, Dejing Dou. Explaining Information Flow Inside Vision Transformers Using Markov Chain. 
        In *Neurips 2021 XAI4Debugging Workshop*. https://openreview.net/forum?id=TT-cf6QSDaQ

    """

    def __init__(self,
                 paddle_model,
                 use_cuda=None,
                 device='gpu:0') -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        Interpreter.__init__(self, paddle_model, device, use_cuda)
        self.paddle_prepared = False

    def interpret(self,
                  inputs,
                  start_layer=4,
                  steps=20,
                  label=None,
                  resize_to=224, 
                  crop_to=None,
                  visual=True,
                  save_path=None):
        """
        Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            start_layer (int, optional): Compute the state from the start layer. Default: 4
            steps (int, optional): number of steps in the Riemman approximation of the integral. Default: 50
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. 
                If None, the most likely label for each image will be used. Default: None
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        Returns:
            [numpy.ndarray]: interpretations/heatmap for images
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)
        bsz = len(data)  # batch size
        save_path = preprocess_save_path(save_path, bsz)

        if not self.paddle_prepared:
            self._paddle_prepare()

        attns, _, preds = self.predict_fn(data)
        assert start_layer < len(attns), "start_layer should be in the range of [0, num_block-1]"

        if label is None:
            label = preds

        b, h, s, _ = attns[0].shape
        num_blocks = len(attns)
        states = np.mean(attns[-1], axis=1)[:, 0, :].reshape((b, 1, s))
        for i in range(start_layer, num_blocks-1)[::-1]:
            attn = np.mean(attns[i], 1)
            states_ = states
            states = states @ attn
            states += states_

        total_gradients = np.zeros((b, h, s, s))
        for alpha in np.linspace(0, 1, steps):
            # forward propagation
            data_scaled = data * alpha
            _, gradients, _ = self.predict_fn(data_scaled, label=label)

            total_gradients += gradients

        W_state = np.mean((total_gradients / steps).clip(min=0), axis=1)[:, 0, :].reshape((b, 1, s))

        tam_explanation = (states * W_state)[:, 0, 1:].reshape((-1, 14, 14))

        # visualization and save image.
        for i in range(bsz):
            vis_explanation = explanation_to_vis(imgs[i], tam_explanation[i], style='overlay_heatmap')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        return tam_explanation

    def _paddle_prepare(self, predict_fn=None):
        if predict_fn is None:
            paddle.set_device(self.device)
            # to get gradients, the ``train`` mode must be set.
            # we cannot set v.training = False for the same reason.
            self.paddle_model.train()

            for n, v in self.paddle_model.named_sublayers():
                if "batchnorm" in v.__class__.__name__.lower():
                    v._use_global_stats = True
                if "dropout" in v.__class__.__name__.lower():
                    v.p = 0

                # Report issues or pull requests if more layers need to be added.

            def predict_fn(data, label=None):
                data = paddle.to_tensor(data)
                data.stop_gradient = False

                attns = []
                def hook(layer, input, output):
                    attns.append(output)

                hooks = []
                for n, v in self.paddle_model.named_sublayers():
                    if re.match('^blocks.*.attn.attn_drop$', n):
                        h = v.register_forward_post_hook(hook)
                        hooks.append(h)
                out = self.paddle_model(data)
                for h in hooks:
                    h.remove()

                out = paddle.nn.functional.softmax(out, axis=1)
                preds = paddle.argmax(out, axis=1)
                if label is None:
                    label = preds.numpy()

                label_onehot = paddle.nn.functional.one_hot(
                    paddle.to_tensor(label), num_classes=out.shape[1])
                target = paddle.sum(out * label_onehot, axis=1)
                target.backward()
                gradients = attns[-1].grad
                target.clear_gradient()
                if isinstance(gradients, paddle.Tensor):
                    gradients = gradients.numpy()

                a = []
                for attn in attns:
                    a.append(attn.numpy())

                return a, gradients, label

        self.predict_fn = predict_fn
        self.paddle_prepared = True



"""
Copyright (c) 2016, Marco Tulio Correia Ribeiro
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
The code in this file (_lime_base.py) is largely simplified and modified from https://github.com/marcotcr/lime.
"""

import numpy as np
import sklearn
import sklearn.preprocessing
from skimage.color import gray2rgb
from sklearn.linear_model import Ridge
from sklearn.utils import check_random_state
from sklearn.preprocessing import normalize
from sklearn.metrics import r2_score, pairwise_distances

import copy
from functools import partial
from skimage.segmentation import quickshift
from skimage.measure import regionprops
from tqdm import tqdm


class LimeBase(object):
    """
    Class for learning a locally linear sparse model from perturbed data
    """

    def __init__(self,
                 kernel_width=0.25,
                 kernel=None,
                 verbose=False,
                 random_state=None):
        """Init function

        """

        if kernel is None:

            def kernel(d, kernel_width):
                return np.sqrt(np.exp(-(d**2) / kernel_width**2))

        kernel_fn = partial(kernel, kernel_width=kernel_width)

        self.kernel_fn = kernel_fn
        self.verbose = verbose
        self.random_state = check_random_state(random_state)

    def _fitting_data(self,
                      neighborhood_data,
                      neighborhood_labels,
                      distances,
                      label,
                      ordered=True,
                      model_regressor=None):

        weights = self.kernel_fn(distances)
        labels_column = neighborhood_labels[:, label]
        used_features = np.array(range(neighborhood_data.shape[1]))

        if model_regressor is None:
            model_regressor = Ridge(
                alpha=0,
                fit_intercept=True,
                normalize=True,
                random_state=self.random_state)
        easy_model = model_regressor
        easy_model.fit(neighborhood_data, labels_column, sample_weight=weights)
        # R^2 between easy_model.predict(X) and y.
        prediction_score = easy_model.score(
            neighborhood_data, labels_column, sample_weight=weights)
        local_pred = easy_model.predict(neighborhood_data[0, used_features]
                                        .reshape(1, -1))

        # if self.verbose:
        #     print('Intercept', easy_model.intercept_)
        #     print(
        #         'Prediction_local',
        #         local_pred, )
        #     print('Right:', neighborhood_labels[0, label])

        if ordered:
            return (easy_model.intercept_, sorted(
                zip(used_features, easy_model.coef_),
                key=lambda x: np.abs(x[1]),
                reverse=True), prediction_score, local_pred)
        else:
            return (easy_model.intercept_,
                    list(zip(used_features, easy_model.coef_)),
                    prediction_score, local_pred)

    def _data_labels(self, image, segments, classifier_fn, num_samples,
                     batch_size, hide_color, distance_metric):
        """Generates images and predictions in the neighborhood of this image.

        Args:
            image: 3d numpy array, the image
            fudged_image: 3d numpy array, image to replace original image when
                superpixel is turned off
            segments: segmentation of the image
            classifier_fn: function that takes a list of images and returns a
                matrix of prediction probabilities
            num_samples: size of the neighborhood to learn the linear model
            batch_size: classifier_fn will be called on batches of this size.

        Returns:
            A tuple (data, labels), where:
                data: dense num_samples * num_superpixels
                labels: prediction probabilities matrix
        """

        fudged_image = image.copy()
        if hide_color is None:
            # if no hide_color, use the mean
            for x in np.unique(segments):
                mx = np.mean(image[segments == x], axis=0)
                fudged_image[segments == x] = mx
        # elif hide_color == 'avg_from_neighbor':
        #     from scipy.spatial.distance import cdist

        #     n_features = np.unique(segments).shape[0]
        #     regions = regionprops(segments + 1)
        #     centroids = np.zeros((n_features, 2))
        #     for i, x in enumerate(regions):
        #         centroids[i] = np.array(x.centroid)

        #     d = cdist(centroids, centroids, 'sqeuclidean')

        #     for x in np.unique(segments):
        #         a = [image[segments == i] for i in np.argsort(d[x])[1:6]]
        #         mx = np.mean(np.concatenate(a), axis=0)
        #         fudged_image[segments == x] = mx
        else:
            fudged_image[:] = 0

        n_features = np.unique(segments).shape[0]
        data = self.random_state.randint(0, 2, num_samples * n_features) \
            .reshape((num_samples, n_features))
        labels = []
        data[0, :] = 1
        imgs = []
        for row in tqdm(data, leave=True, position=0):
            temp = copy.deepcopy(image)
            zeros = np.where(row == 0)[0]
            mask = np.zeros(segments.shape).astype(bool)
            for z in zeros:
                mask[segments == z] = True
            temp[mask] = fudged_image[mask]
            imgs.append(temp)
            if len(imgs) == batch_size:
                preds = classifier_fn(np.array(imgs))
                if isinstance(preds, list):
                    preds = preds[0]

                labels.extend(preds)
                imgs = []
        if len(imgs) > 0:
            preds = classifier_fn(np.array(imgs))
            labels.extend(preds)

        distances = sklearn.metrics.pairwise_distances(
            data, data[0].reshape(1, -1), metric=distance_metric).ravel()

        return data, np.array(labels), distances

    def _fitting_data_with_prior(self,
                                 neighborhood_data,
                                 neighborhood_labels,
                                 distances,
                                 label,
                                 prior=None,
                                 prior_scale=1.0,
                                 reg_force=1.0):

        weights = self.kernel_fn(distances)
        labels_column = neighborhood_labels[:, label]
        used_features = np.array(range(neighborhood_data.shape[1]))

        # use this regressor just for creating an instance of estimator.
        model_regressor = Ridge(
            alpha=0, fit_intercept=True, random_state=self.random_state)
        easy_model = model_regressor

        X = np.float32(neighborhood_data[:, used_features])
        y = np.copy(labels_column)

        # pre-process
        X_offset = np.average(X, axis=0, weights=weights)
        X -= X_offset
        X, X_scale = normalize(X, axis=0, copy=True, return_norm=True)

        y_offset = np.average(y, axis=0, weights=weights)
        y -= y_offset
        y = np.reshape(y, (-1, 1))
        y, y_scale = normalize(y, axis=0, copy=True, return_norm=True)
        y = np.reshape(y, (-1))

        if weights is not None:
            # rescale by weights
            sample_weights = np.diag(np.sqrt(weights))
            X = np.dot(sample_weights, X)
            y = np.dot(sample_weights, y)

        if prior is not None:
            w0 = np.zeros(len(used_features)) if prior is None else prior
            w0 = np.array(w0)
            if np.sum(np.abs(w0)) > 0.0:
                w0 = w0 / np.sum(np.abs(w0))
        else:
            w0 = np.zeros(len(used_features))

        I = np.identity(len(used_features))

        beta = reg_force
        tau = prior_scale
        w = np.dot(
            np.linalg.inv(np.dot(X.T, X) + beta * I),
            np.dot(X.T, y) + beta * tau * w0)

        easy_model.coef_ = w
        easy_model.intercept_ = 0

        y_pred = np.dot(X, w) * y_scale + y_offset
        predict_r2_score = r2_score(
            labels_column,
            y_pred,
            sample_weight=weights,
            multioutput='variance_weighted')

        # if self.verbose:
        #     print('Intercept', easy_model.intercept_)
        #     print('Right:', neighborhood_labels[0, label])
        return (easy_model.intercept_, sorted(
            zip(used_features, easy_model.coef_),
            key=lambda x: np.abs(x[1]),
            reverse=True), predict_r2_score, 0)

    def interpret_instance(self,
                           image,
                           classifier_fn,
                           interpret_labels=(1, ),
                           num_samples=1000,
                           batch_size=10,
                           hide_color=None,
                           distance_metric='cosine',
                           model_regressor=None,
                           segments=None,
                           prior=None,
                           reg_force=1.0):
        """
        Generates interpretations for a prediction.
        """
        if len(image.shape) == 2:
            image = gray2rgb(image)

        if segments is None:
            segments = compute_segments(image)

        self.segments = segments

        data, labels, distances = self._data_labels(
            image, segments, classifier_fn, num_samples, batch_size,
            hide_color, distance_metric)

        lime_weights = {}
        prediction_scores = {}
        for l in interpret_labels:
            if prior is None:
                (_, lime_weights[l], prediction_scores[l],
                 _) = self._fitting_data(data, labels, distances, l, True,
                                         model_regressor)
            else:
                (_, lime_weights[l], prediction_scores[l],
                 _) = self._fitting_data_with_prior(
                     data, labels, distances, l, prior, reg_force=reg_force)

        return lime_weights, prediction_scores

    def interpret_instance_text(self,
                                model_inputs,
                                classifier_fn,
                                interpret_labels,
                                num_samples,
                                batch_size,
                                unk_id,
                                pad_id,
                                distance_metric='cosine',
                                model_regressor=None,
                                prior=None,
                                reg_force=1.0):
        """
        Generates interpretations for a prediction.
        """
        #word_ids = np.array(model_inputs[0])
        #if len(word_ids.shape) > 1:
        #    word_ids = word_ids[0]
        data, labels, distances = self._data_labels_text(
            model_inputs, classifier_fn, num_samples, batch_size,
            distance_metric, unk_id, pad_id)
        lime_weights = {}
        prediction_scores = {}
        for l in interpret_labels:
            if prior is None:
                (_, lime_weights[l], prediction_scores[l],
                 _) = self._fitting_data(data, labels, distances, l, False,
                                         model_regressor)
            else:
                (_, lime_weights[l], prediction_scores[l],
                 _) = self._fitting_data_with_prior(
                     data, labels, distances, l, prior, reg_force=reg_force)

        return lime_weights, prediction_scores

    def _data_labels_text(self, model_inputs, classifier_fn, num_samples,
                          batch_size, distance_metric, unk_id, pad_id):
        import paddle
        word_ids = model_inputs[0]
        if not isinstance(word_ids, np.ndarray):
            word_ids = word_ids.numpy()
        ori_shape = word_ids.shape
        word_ids = word_ids.reshape((np.prod(ori_shape), ))
        if pad_id is None:
            n_features = len(word_ids)
        else:
            pad_locs = np.where(word_ids == pad_id)[0]
            n_features = word_ids.shape[-1] if len(pad_locs) == 0 else min(
                pad_locs)
        data = self.random_state.randint(0, 2, num_samples * n_features) \
            .reshape((num_samples, n_features))
        labels = []
        data[0, :] = 1
        samples = []
        for row in data:
            temp = copy.deepcopy(word_ids)
            zeros = np.where(row == 0)[0]
            for z in zeros:
                temp[z] = unk_id
            samples.append(temp.reshape(ori_shape).tolist()[0])
            if len(samples) == batch_size:
                pred_inputs = (np.array(samples), ) + tuple([
                    np.repeat(
                        inp, batch_size, axis=0) for inp in model_inputs[1:]
                ])
                preds = classifier_fn(*pred_inputs).tolist()
                labels.extend(preds)
                samples = []
        if len(samples) > 0:
            pred_inputs = (np.array(samples), ) + tuple([
                np.repeat(
                    inp, len(samples), axis=0) for inp in model_inputs[1:]
            ])
            preds = classifier_fn(*pred_inputs).tolist()
            labels.extend(preds)

        distances = sklearn.metrics.pairwise_distances(
            data, data[0].reshape(1, -1), metric=distance_metric).ravel()
        return data, np.array(labels), distances


def compute_segments(image):
    assert len(image.shape) == 3 and image.shape[
        -1] == 3, "Shape Error when computing superpixels. " + str(image.shape)
    segments = quickshift(image, sigma=1)
    return segments



import paddle
import unittest
import numpy as np
import interpretdl as it
from paddle.utils.download import get_weights_path_from_url

from tutorials.assets.vision_transformer import ViT_base_patch16_224
from tests.utils import assert_arrays_almost_equal


class TestTAM(unittest.TestCase):

    def set_paddle_model(self):
        paddle_model = ViT_base_patch16_224()
        model_path = get_weights_path_from_url(
            'https://paddle-imagenet-models-name.bj.bcebos.com/dygraph/ViT_base_patch16_224_pretrained.pdparams'
        )
        paddle_model.set_dict(paddle.load(model_path))
        return paddle_model

    def test_shape(self):
        paddle_model = self.set_paddle_model()
        img_path = 'tutorials/assets/catdog.png'
        algo = it.TAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, steps=2, resize_to=256, crop_to=224, visual=False)
        result = np.array([*exp.shape])
        assert_arrays_almost_equal(self, result, np.array([1, 14, 14]))

    def test_cv_multiple_inputs(self):
        paddle_model = self.set_paddle_model()

        img_path = ['tutorials/assets/catdog.png', 'tutorials/assets/catdog.png']
        algo = it.TAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, steps=2, resize_to=256, crop_to=224, visual=False)
        result = np.array([*exp.shape])
        assert_arrays_almost_equal(self, result, np.array([2, 14, 14]))

    def test_algo(self):
        paddle_model = self.set_paddle_model()

        np.random.seed(42)
        img_path = np.random.randint(0, 255, size=(1, 224, 224, 3), dtype=np.uint8)
        algo = it.TAMInterpreter(paddle_model, device='cpu')
        exp = algo.interpret(img_path, steps=2, visual=False)
        result = np.array([exp.mean(), exp.std(), exp.min(), exp.max()])
        desired = np.array([0.00122923, 0.00068821, 0.00013711, 0.00403947])

        assert_arrays_almost_equal(self, result, desired)        


if __name__ == '__main__':
    unittest.main()




import numpy as np
from tqdm import tqdm
from .abc_interpreter import InputGradientInterpreter, Interpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image


class IntGradCVInterpreter(InputGradientInterpreter):
    """
    Integrated Gradients Interpreter for CV tasks.

    More details regarding the Integrated Gradients method can be found in the original paper:
    https://arxiv.org/abs/1703.01365
    """

    def __init__(
        self,
        paddle_model: callable,
        use_cuda: bool=True,
        device: str='gpu:0'
    ):
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        InputGradientInterpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(
        self,
        inputs: str or list(str) or np.ndarray,
        labels: list or tuple or np.ndarray or None=None,
        baselines: np.ndarray or None=None,
        steps: int=50,
        num_random_trials: int=10,
        resize_to=224, 
        crop_to=None,
        visual: bool=True,
        save_path=None
    ):
        """Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. If None, the most likely label for each image will be used. Default: None            baseline (numpy.ndarray, optional): The baseline input. If None, all zeros will be used. Default: None
            baselines (numpy.ndarray or None, optional): The baseline images to compare with. It should have the same shape as images and same length as the number of images.
                                                        If None, the baselines of all zeros will be used. Default: None.
            steps (int, optional): number of steps in the Riemman approximation of the integral. Default: 50
            num_random_trials (int, optional): number of random initializations to take average in the end. Default: 10
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        Returns:
            [numpy.ndarray]: interpretations/gradients for images
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)

        self.data_type = np.array(data).dtype
        self.input_type = type(data)

        self._build_predict_fn(gradient_of='probability')

        if baselines is None:
            num_random_trials = 1
            self.baselines = np.zeros(
                (num_random_trials, ) + data.shape, dtype=self.data_type)
        elif baselines == 'random':
            self.baselines = np.random.normal(
                size=(num_random_trials, ) + data.shape).astype(self.data_type)
        else:
            self.baselines = baselines
        bsz = len(data)

        # obtain the labels (and initialization).
        if labels is None:
            gradients, preds = self.predict_fn(data, labels)
            labels = preds
        labels = np.array(labels).reshape((bsz, ))

        # IntGrad.
        gradients_list = []
        with tqdm(total=num_random_trials * steps, leave=True, position=0) as pbar:
            for i in range(num_random_trials):
                total_gradients = np.zeros_like(data)
                for alpha in np.linspace(0, 1, steps):
                    data_scaled = data * alpha + self.baselines[i] * (1 - alpha)
                    gradients, _ = self.predict_fn(data_scaled, labels)
                    total_gradients += gradients
                    pbar.update(1)

                ig_gradients = total_gradients * (data - self.baselines[i]) / steps
                gradients_list.append(ig_gradients)
        avg_gradients = np.average(np.array(gradients_list), axis=0)

        # visualization and save image.
        if save_path is None and not visual:
            # no need to visualize or save explanation results.
            pass
        else:
            save_path = preprocess_save_path(save_path, bsz)
            for i in range(bsz):
                vis_explanation = explanation_to_vis(imgs[i], np.abs(avg_gradients[i]).sum(0), style='overlay_grayscale')
                if visual:
                    show_vis_explanation(vis_explanation)
                if save_path[i] is not None:
                    save_image(save_path[i], vis_explanation)

        # intermediate results, for possible further usages.
        self.labels = labels

        return avg_gradients


class IntGradNLPInterpreter(Interpreter):
    """
    Integrated Gradients Interpreter for NLP tasks.

    More details regarding the Integrated Gradients method can be found in the original paper:
    https://arxiv.org/abs/1703.01365
    """

    def __init__(self, paddle_model, use_cuda=True, device='gpu:0') -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        Interpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(self,
                  data,
                  labels=None,
                  steps=50,
                  embedding_name='word_embeddings',
                  return_pred=True):
        """Main function of the interpreter.

        Args:
            data ([type]): [description]
            labels ([type], optional): The target label to analyze. If None, the most likely label will be used. Default: None.
            steps (int, optional): number of steps in the Riemman approximation of the integral. Default: 50
            embedding_name (str, optional): name of the embedding layer at which the steps will be applied. 
                Defaults to 'word_embeddings'. The correct name of embedding can be found through ``print(model)``.
            return_pred (bool, optional): Whether or not to return predicted labels and probabilities. 
                If True, a tuple of predicted labels, probabilities, and interpretations will be returned.
                There are useful for visualization. Else, only interpretations will be returned. Default: True.

        Returns:
            [numpy.ndarray or tuple]: interpretations for each word or a tuple of predicted labels, probabilities, and interpretations.
        """

        self._build_predict_fn(embedding_name=embedding_name, gradient_of='probability')

        if isinstance(data, tuple):
            bs = data[0].shape[0]
        else:
            bs = data.shape[0]

        gradients, labels, data_out, probas = self.predict_fn(data, labels, None)
        labels = labels.reshape((bs, ))
        total_gradients = np.zeros_like(gradients)
        for alpha in np.linspace(0, 1, steps):
            gradients, _, _, _ = self.predict_fn(data, labels, alpha)
            total_gradients += gradients
        
        ig_gradients = total_gradients * data_out / steps
        ig_gradients = np.sum(ig_gradients, axis=-1)

        if return_pred:
            return labels, probas.numpy(), ig_gradients
        
        # Visualization is currently not supported here.
        # See the tutorial for more information:
        # https://github.com/PaddlePaddle/InterpretDL/blob/master/tutorials/ernie-2.0-en-sst-2-tutorials.ipynb

        return ig_gradients

    def _build_predict_fn(self, rebuild=False, embedding_name='word_embeddings', gradient_of='probability'):
        
        if self.predict_fn is not None:
            assert callable(self.predict_fn), "predict_fn is predefined before, but is not callable." \
                "Check it again."
            return
        
        import paddle
        if self.predict_fn is None or rebuild:
            assert gradient_of in ['loss', 'logit', 'probability']

            if not paddle.is_compiled_with_cuda() and self.device[:3] == 'gpu':
                print("Paddle is not installed with GPU support. Change to CPU version now.")
                self.device = 'cpu'

            # set device. self.device is one of ['cpu', 'gpu:0', 'gpu:1', ...]
            paddle.set_device(self.device)

            # to get gradients, the ``train`` mode must be set.
            self.paddle_model.train()
            
            # later version will be simplied.
            for n, v in self.paddle_model.named_sublayers():
                if "batchnorm" in v.__class__.__name__.lower():
                    v._use_global_stats = True
                if "dropout" in v.__class__.__name__.lower():
                    v.p = 0
                    
            def predict_fn(data, labels, noise_scale=1.0):
                if isinstance(data, tuple):
                    # NLP models usually have two inputs.
                    bs = data[0].shape[0]
                    data = (paddle.to_tensor(data[0]), paddle.to_tensor(data[1]))
                else:
                    bs = data.shape[0]
                    data = paddle.to_tensor(data)
                    
                assert labels is None or \
                    (isinstance(labels, (list, np.ndarray)) and len(labels) == bs)
            
                target_feature_map = []
                def hook(layer, input, output):
                    if noise_scale is not None:
                        output = noise_scale * output
                    target_feature_map.append(output)
                    return output
                hooks = []
                for name, v in self.paddle_model.named_sublayers():
                    if embedding_name in name:
                        h = v.register_forward_post_hook(hook)
                        hooks.append(h)
                        
                if isinstance(data, tuple):
                    logits = self.paddle_model(*data)  # get logits, [bs, num_c]
                else:
                    logits = self.paddle_model(data)  # get logits, [bs, num_c]
                    
                for h in hooks:
                    h.remove()
                    
                probas = paddle.nn.functional.softmax(logits, axis=1)  # get probabilities.
                preds = paddle.argmax(probas, axis=1)  # get predictions.
                if labels is None:
                    labels = preds.numpy()  # label is an integer.
                
                if gradient_of == 'loss':
                    # loss
                    loss = paddle.nn.functional.cross_entropy(
                        logits, paddle.to_tensor(labels), reduction='sum'
                    )
                else:
                    # logits or probas
                    labels = np.array(labels).reshape((bs, ))
                    labels_onehot = paddle.nn.functional.one_hot(
                        paddle.to_tensor(labels), num_classes=probas.shape[1]
                    )
                    if gradient_of == 'logit':
                        loss = paddle.sum(logits * labels_onehot, axis=1)
                    else:
                        loss = paddle.sum(probas * labels_onehot, axis=1)

                loss.backward()
                gradients = target_feature_map[0].grad  # get gradients of "embedding".
                loss.clear_gradient()

                if isinstance(gradients, paddle.Tensor):
                    gradients = gradients.numpy()
                return gradients, labels, target_feature_map[0].numpy(), probas

        self.predict_fn = predict_fn



from .abc_interpreter import Interpreter, InputGradientInterpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image

import numpy as np


class GradShapCVInterpreter(InputGradientInterpreter):
    """
    Gradient SHAP Interpreter for CV tasks.

    More details regarding the GradShap method can be found in the original paper:
    http://papers.nips.cc/paper/7062-a-unified-approach-to-interpreting-model-predictions
    """

    def __init__(
        self, 
        paddle_model: callable,
        use_cuda: bool=None,
        device: str='gpu:0'
    ):
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        InputGradientInterpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(
        self,
        inputs: str or list(str) or np.ndarray,
        labels: list or np.ndarray=None,
        baselines: np.ndarray=None,
        n_samples: int=5,
        noise_amount: float=0.1,
        resize_to=224, 
        crop_to=None,
        visual: bool=True,
        save_path=None
    ) -> np.ndarray:
        """
        Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. If None, the most likely label for each image will be used. Default: None            baseline (numpy.ndarray, optional): The baseline input. If None, all zeros will be used. Default: None
            baselines (numpy.ndarray or None, optional): The baseline images to compare with. It should have the same shape as images and same length as the number of images.
                                                        If None, the baselines of all zeros will be used. Default: None.
            n_samples (int, optional): The number of randomly generated samples. Default: 5.
            noise_amount (float, optional): Noise level of added noise to each image.
                                            The std of Guassian random noise is noise_amount * (x_max - x_min). Default: 0.1
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True.
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        Returns:
            [numpy.ndarray]: interpretations for images
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)
        bsz = len(data)
        self.data_type = np.array(data).dtype

        self._build_predict_fn(gradient_of='probability')

        if labels is None:
            _, labels = self.predict_fn(data, labels)

        def add_noise_to_inputs(data):
            max_axis = tuple(np.arange(1, data.ndim))
            stds = noise_amount * (
                np.max(data, axis=max_axis) - np.min(data, axis=max_axis))
            noise = np.concatenate([
                np.random.normal(0.0, stds[j], (n_samples, ) + tuple(d.shape))
                for j, d in enumerate(data)
            ]).astype(self.data_type)
            repeated_data = np.repeat(data, (n_samples, ) * len(data), axis=0)
            return repeated_data + noise
        
        data_with_noise = add_noise_to_inputs(data)

        if baselines is None:
            baselines = np.zeros_like(data)
        baselines = np.repeat(baselines, (n_samples, ) * bsz, axis=0)

        labels = np.array(labels).reshape(
            (bsz, 1))  #.repeat(n_samples, axis=0)
        labels = np.repeat(labels, (n_samples, ) * bsz, axis=0)

        rand_scales = np.random.uniform(
            0.0, 1.0, (bsz * n_samples, 1)).astype(self.data_type)

        input_baseline_points = np.array([
            d * r + b * (1 - r)
            for d, r, b in zip(data_with_noise, rand_scales, baselines)
        ])

        gradients, _ = self.predict_fn(input_baseline_points, labels)

        input_baseline_diff = data_with_noise - baselines
        explanations = gradients * input_baseline_diff

        explanations = np.concatenate([
            np.mean(
                explanations[i * n_samples:(i + 1) * n_samples],
                axis=0,
                keepdims=True) for i in range(bsz)
        ])

        # visualization and save image.
        save_path = preprocess_save_path(save_path, bsz)
        for i in range(bsz):
            vis_explanation = explanation_to_vis(imgs[i], np.abs(explanations[i]).sum(0), style='overlay_grayscale')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        return explanations


class GradShapNLPInterpreter(Interpreter):
    """
    Gradient SHAP Interpreter for NLP tasks.

    More details regarding the GradShap method can be found in the original paper:
    http://papers.nips.cc/paper/7062-a-unified-approach-to-interpreting-model-predictions
    """

    def __init__(
        self, paddle_model: callable, use_cuda: bool=True, device: str='gpu:0'
    ) -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        Interpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(
        self,
        data,
        labels=None,
        n_samples=5,
        noise_amount=0.1,
        embedding_name='word_embeddings',
        return_pred=True
    ):
        """Main function of the interpreter.

        Args:
            data ([type]): The inputs to the NLP model.
            labels ([type], optional): The target label to analyze. If None, the most likely label will be used. Default: None.
            n_samples (int, optional): [description]. Defaults to 5.
            noise_amount (float, optional): Noise level of added noise to the embeddings. 
                The std of Guassian random noise is ``noise_amount * embedding.mean() * (x_max - x_min)``. Default: 0.1
            embedding_name (str, optional): name of the embedding layer at which the noises will be applied. 
                Defaults to 'word_embeddings'. The correct name of embedding can be found through ``print(model)``.
            return_pred (bool, optional): Whether or not to return predicted labels and probabilities. 
                If True, a tuple of predicted labels, probabilities, and interpretations will be returned.
                There are useful for visualization. Else, only interpretations will be returned. Default: True.

        Returns:
            [type]: [description]
        """

        self._build_predict_fn(embedding_name=embedding_name)

        if isinstance(data, tuple):
            bs = data[0].shape[0]
        else:
            bs = data.shape[0]

        gradients, labels, data_out, probas = self.predict_fn(data, labels, None)
        labels = labels.reshape((bs, ))
        total_gradients = np.zeros_like(gradients)

        rand_scales = np.random.uniform(0.0, 1.0, (n_samples, ))

        for scale in rand_scales:
            gradients, _, _, _ = self.predict_fn(data, labels, scale, noise_amount)
            total_gradients += np.array(gradients)

        interpretations = total_gradients * data_out / n_samples
        interpretations = np.sum(interpretations, axis=-1)

        # Visualization is currently not supported here.
        # See the tutorial for more information:
        # https://github.com/PaddlePaddle/InterpretDL/blob/master/tutorials/ernie-2.0-en-sst-2-tutorials.ipynb
        if return_pred:
            return labels, probas.numpy(), interpretations

        return interpretations

    def _build_predict_fn(self, rebuild=False, embedding_name='word_embeddings'):
        
        if self.predict_fn is not None:
            assert callable(self.predict_fn), "predict_fn is predefined before, but is not callable." \
                "Check it again."
            return
        
        import paddle
        if self.predict_fn is None or rebuild:

            if not paddle.is_compiled_with_cuda() and self.device[:3] == 'gpu':
                print("Paddle is not installed with GPU support. Change to CPU version now.")
                self.device = 'cpu'

            # set device. self.device is one of ['cpu', 'gpu:0', 'gpu:1', ...]
            paddle.set_device(self.device)

            # to get gradients, the ``train`` mode must be set.
            self.paddle_model.train()
            
            # later version will be simplied.
            for n, v in self.paddle_model.named_sublayers():
                # if "batchnorm" in v.__class__.__name__.lower():
                #     v._use_global_stats = True
                if "dropout" in v.__class__.__name__.lower():
                    v.p = 0
                    
            def predict_fn(data, labels, scale=None, noise_amount=None):
                if isinstance(data, tuple):
                    # NLP models usually have two inputs.
                    bs = data[0].shape[0]
                    data = (paddle.to_tensor(data[0]), paddle.to_tensor(data[1]))
                else:
                    bs = data.shape[0]
                    data = paddle.to_tensor(data)
                    
                assert labels is None or \
                    (isinstance(labels, (list, np.ndarray)) and len(labels) == bs)
            
                target_feature_map = []
                def hook(layer, input, output):
                    if noise_amount is not None:
                        bias = paddle.normal(std=noise_amount * output.mean(), shape=output.shape)
                        output = output + bias
                    if scale is not None:
                        output = scale * output
                    target_feature_map.append(output)
                    return output
                hooks = []
                for name, v in self.paddle_model.named_sublayers():
                    if embedding_name in name:
                        h = v.register_forward_post_hook(hook)
                        hooks.append(h)
                        
                if isinstance(data, tuple):
                    logits = self.paddle_model(*data)  # get logits, [bs, num_c]
                else:
                    logits = self.paddle_model(data)  # get logits, [bs, num_c]
                    
                for h in hooks:
                    h.remove()
                    
                probas = paddle.nn.functional.softmax(logits, axis=1)  # get probabilities.
                preds = paddle.argmax(probas, axis=1)  # get predictions.
                if labels is None:
                    labels = preds.numpy()  # label is an integer.
                
                
                # logits or probas
                labels = np.array(labels).reshape((bs, ))
                labels_onehot = paddle.nn.functional.one_hot(
                    paddle.to_tensor(labels), num_classes=probas.shape[1]
                )
                loss = paddle.sum(probas * labels_onehot, axis=1)

                loss.backward()
                gradients = target_feature_map[0].grad  # get gradients of "embedding".
                loss.clear_gradient()

                if isinstance(gradients, paddle.Tensor):
                    gradients = gradients.numpy()
                return gradients, labels, target_feature_map[0].numpy(), probas

        self.predict_fn = predict_fn


#!/usr/bin/env python3

import os
from setuptools import setup, find_packages
import interpretdl

with open("requirements.txt") as fin:
    REQUIRED_PACKAGES = fin.read()

# get key package details from py_pkg/__version__.py
about = {
    '__title__': 'interpretdl',
    '__description__': 'interpretation of deep learning models',
    '__version__': interpretdl.__version__
}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
# with open(os.path.join(here, 'py_pkg', '__version__.py')) as f:
#     exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
#with open('README.md', 'r') as f:
#    readme = f.read()

# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
long_description = 'Home-page: https://github.com/PaddlePaddle/InterpretDL \
Author: Baidu-BDL \
Author-email: autodl@baidu.com \
License: Apache 2.0 \
Description: InterpretDL, short for interpretation of deep learning models, is a model interpretation toolkit for PaddlePaddle models.'

setup(
    name=about['__title__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/plain',
    version=about['__version__'],
    #url=about['__url__'],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7.*",
    install_requires=REQUIRED_PACKAGES,
    license='Apache 2.0',
    zip_safe=False,
    entry_points={
        'console_scripts': ['interpretdl=interpretdl.command:main'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='interpreters for Paddle models')



import unittest
from paddle.vision.models import mobilenet_v2
import numpy as np
import os
import interpretdl as it
from tests.utils import assert_arrays_almost_equal


class TestLIMENLP(unittest.TestCase):

    def prepare(self):
        import paddle
        import paddlenlp
        from paddlenlp.data import Stack, Tuple, Pad
        from paddlenlp.transformers import ErnieForSequenceClassification, ErnieTokenizer
        from tutorials.assets.utils import convert_example, aggregate_subwords_and_importances
        
        MODEL_NAME = "ernie-2.0-en"
        model = ErnieForSequenceClassification.from_pretrained(MODEL_NAME, num_classes=2)
        tokenizer = ErnieTokenizer.from_pretrained(MODEL_NAME)

        self.paddle_model = model
        self.tokenizer = tokenizer

        def preprocess_fn(data):
            examples = []
            
            if not isinstance(data, list):
                data = [data]
            
            for text in data:
                input_ids, segment_ids = convert_example(
                    text,
                    tokenizer,
                    max_seq_length=128,
                    is_test=True
                )
                examples.append((input_ids, segment_ids))

            batchify_fn = lambda samples, fn=Tuple(
                Pad(axis=0, pad_val=tokenizer.pad_token_id),  # input id
                Pad(axis=0, pad_val=tokenizer.pad_token_id),  # segment id
            ): fn(samples)
            
            input_ids, segment_ids = batchify_fn(examples)
            return paddle.to_tensor(input_ids, stop_gradient=False), paddle.to_tensor(segment_ids, stop_gradient=False)
        self.preprocess_fn = preprocess_fn

    def test_lime_intgrad_gradshap_nlp(self):
        self.prepare()

        reviews = [
            "it 's a charming and often affecting journey . ",
        ]
        data = [ {"text": r} for r in reviews]

        algo = it.LIMENLPInterpreter(self.paddle_model, device='cpu')
        for i, review in enumerate(data):
            pred_class, pred_prob, lime_weights = algo.interpret(
                review,
                self.preprocess_fn,
                num_samples=22,
                batch_size=20,
                unk_id=self.tokenizer.convert_tokens_to_ids('[UNK]'),
                pad_id=self.tokenizer.convert_tokens_to_ids('[PAD]'),
                return_pred=True)

        algo = it.IntGradNLPInterpreter(self.paddle_model, device='cpu')
        pred_labels, pred_probs, avg_gradients = algo.interpret(
            self.preprocess_fn(data),
            steps=2,
            return_pred=True)

        algo = it.GradShapNLPInterpreter(self.paddle_model, device='cpu')

        pred_labels, pred_probs, avg_gradients = algo.interpret(
            self.preprocess_fn(data),
            n_samples=2,
            noise_amount=0.1,
            return_pred=True)
            
    def test_normlime(self):
        self.prepare()

        reviews = [
            "it 's a charming and often affecting journey . ",
            'the movie achieves as great an impact by keeping these thoughts hidden as ... ( quills ) did by showing them . '
        ]

        data = [ {"text": r} for r in reviews]
        
        normlime = it.NormLIMENLPInterpreter(self.paddle_model, device='cpu')
        normlime.interpret(
            data, self.preprocess_fn, 20, 20,
            unk_id=self.tokenizer.convert_tokens_to_ids('[UNK]'),
            pad_id=self.tokenizer.convert_tokens_to_ids('[PAD]'),
        )

        os.remove('all_lime_weights.npz')
        os.remove('normlime_weights.npy')



import paddle
import numpy as np

from .abc_interpreter import Interpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image


class LRPCVInterpreter(Interpreter):
    """
    Layer-wise Relevance Propagation Interpreter for CV tasks.

    More details regarding the Layer-wise Relevance Propagation method can be found in the original paper:
    https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0130140
    """

    def __init__(self,
                 paddle_model,
                 use_cuda=True,
                 device='gpu:0') -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.

        """
        Interpreter.__init__(self, paddle_model, device, use_cuda)
        self.paddle_prepared = False

    def interpret(self,
                  inputs,
                  label=None,
                  resize_to=224, 
                  crop_to=None,
                  visual=True,
                  save_path=None):
        """
        Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. 
                If None, the most likely label for each  image will be used. Default: None
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None
        
        Returns:
            [numpy.ndarray]: interpretations/Relevance map for images.
            
        """
        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)
        
        bsz = len(data)
        save_path = preprocess_save_path(save_path, bsz)

        if not self.paddle_prepared:
            self._paddle_prepare()

        R, output = self.predict_fn(data, label)

        # visualization and save image.
        for i in range(len(imgs)):
            vis_explanation = explanation_to_vis(imgs[i], R[i].squeeze(), style='overlay_grayscale')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        return R

    def _paddle_prepare(self, predict_fn=None):
        if predict_fn is None:
            paddle.set_device(self.device)
            self.paddle_model.eval()

            layer_list = [(n, v) for n, v in self.paddle_model.named_sublayers()]
            num_classes = layer_list[-1][1].weight.shape[1]

            def predict_fn(data, label):
                data = paddle.to_tensor(data, stop_gradient=False)
                output = self.paddle_model(data)

                if label is None:
                    T = output.argmax().numpy()[0]
                else:
                    assert isinstance(label, int), "label should be an integer"
                    assert 0 <= label < num_classes, f"input label is not correct, label should be at [0, {num_classes})"
                    T = label

                T = np.expand_dims(T, 0)
                T = (T[:, np.newaxis] == np.arange(num_classes)) * 1.0
                T = paddle.to_tensor(T).astype('float32')

                R = self.paddle_model.relprop(
                    R=output * T, alpha=1).sum(axis=1, keepdim=True)

                # Check relevance value preserved
                # print("Check relevance value preserved: ")
                # print('Pred logit : ' + str((output * T).sum().cpu().numpy()))
                # print('Relevance Sum : ' + str(R.sum().cpu().numpy()))

                return R.detach().numpy(), output.detach().numpy()

        self.predict_fn = predict_fn
        self.paddle_prepared = True

        return predict_fn




from typing import Any, Callable, List, Tuple, Union
import numpy as np
import paddle

from .lime import LIMECVInterpreter
from ._lime_base import compute_segments
from ._global_prior_base import precompute_global_prior, use_fast_normlime_as_prior
from ..data_processor.readers import images_transform_pipeline, preprocess_image, load_npy_dict_file
from ..data_processor.visualizer import sp_weights_to_image_explanation, overlay_threshold, save_image, show_vis_explanation

class LIMEPriorInterpreter(LIMECVInterpreter):
    """
    LIME Prior Interpreter.
    """

    def __init__(self,
                 paddle_model: Callable,
                 prior_method="none",
                 use_cuda=True,
                 device='gpu:0') -> None:
        """
        Args:
            paddle_model (callable): A paddle model that outputs predictions.
            prior_method: Prior method. Can be chosen from ``{"none", "ridge"}``.
                Defaults to ``"none"``, which is equivalent to LIME.
                If ``"none"``, ``interpret()`` will use zeros as prior;
                Otherwise, the loaded prior will be used.
            use_cuda: Whether to use CUDA. Defaults to ``True``.
        """

        LIMECVInterpreter.__init__(self, paddle_model, use_cuda, device)
        self.prior_method = prior_method
        self.global_weights = None

    def interpreter_init(self,
                         list_file_paths=None,
                         batch_size=0,
                         weights_file_path=None):
        """
        Pre-compute global weights.

        If ``weights_file_path`` is given and has contents, then skip the pre-compute process and
        read the content as the global weights; else the pre-compute process should be performed
        with `list_file_paths`. After the pre-compute process, if `weights_file_path` is given,
        then the prec-computed weights will be saved in this path for skipping the pre-compute
        process in the next usage.
        This is an additional step that LIMEPrior Interpreter has to perform before interpretation.

        Args:
            list_file_paths: List of files used to compute the global prior.
            batch_size: Number of samples to forward each time.
            weights_file_path: Path to load as prior.

        Returns:
            None.

        """

        assert list_file_paths is not None or weights_file_path is not None, \
            "Cannot prepare without anything. "

        self._build_predict_fn(rebuild=True, output='probability')
        def predict_fn_for_lime(_imgs):
            _data = preprocess_image(
                _imgs
            )  # transpose to [N, 3, H, W], scaled to [0.0, 1.0]
            
            output, _ = self.predict_fn(_data, None)
            return output
        self.predict_fn_for_lime = predict_fn_for_lime

        precomputed_weights = load_npy_dict_file(weights_file_path)
        if precomputed_weights is not None:
            self.global_weights = precomputed_weights
        else:
            self.global_weights = precompute_global_prior(
                list_file_paths, self.predict_fn_for_lime, batch_size,
                self.prior_method)
            if weights_file_path is not None and self.global_weights is not None:
                np.save(weights_file_path, self.global_weights)

    def interpret(self,
                  inputs,
                  interpret_class=None,
                  prior_reg_force=1.0,
                  num_samples=1000,
                  batch_size=50,
                  resize_to=256, 
                  crop_to=224,
                  visual=True,
                  save_path=None):
        """
        Note that for LIME prior interpreter, ``interpreter_init()`` needs to be called before calling ``interpret()``.

        Args:
            inputs (str): The input file path.
            interpret_class (int, optional): The index of class to interpret. If None, the most likely label will be used. Default: None
            prior_reg_force (float, optional): The regularization force to apply. Default: 1.0
            num_samples (int, optional): LIME sampling numbers. Larger number of samples usually gives more accurate interpretation. Default: 1000
            batch_size (int, optional): Number of samples to forward each time. Default: 50
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str, optional): The path to save the processed image. If None, the image will not be saved. Default: None

        Returns:
            [dict]: LIME results: {interpret_label_i: weights on features}

        """
        if self.global_weights is None and self.prior_method != "none":
            raise ValueError(
                "The interpreter is not prepared. Call prepare() before interpretation."
            )
        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)

        self._build_predict_fn(output='probability')

        probability, _ = self.predict_fn(data, None)
        # only one example here
        probability = probability[0]

        if interpret_class is None:
            # only interpret top 1 if not provided.
            pred_label = np.argsort(probability)
            interpret_class = pred_label[-1:]
            interpret_class = np.array(interpret_class)
        elif isinstance(interpret_class, list):
            interpret_class = np.array(interpret_class)
        else:
            interpret_class = np.array([interpret_class])
        
        # def predict_fn_for_lime(_imgs):
        #     _data = preprocess_image(
        #         _imgs
        #     )  # transpose to [N, 3, H, W], scaled to [0.0, 1.0]
            
        #     output, _ = self.predict_fn(_data, None)
        #     return output

        # self.predict_fn_for_lime = predict_fn_for_lime

        segments = compute_segments(imgs[0])
        if self.prior_method == "none":
            prior = np.zeros(len(np.unique(segments)))
        else:
            prior = use_fast_normlime_as_prior(imgs, segments,
                                               interpret_class[0],
                                               self.global_weights)

        lime_weights, r2_scores = self.lime_base.interpret_instance(
            imgs[0],
            self.predict_fn_for_lime,
            interpret_class,
            num_samples=num_samples,
            batch_size=batch_size,
            prior=prior,
            reg_force=prior_reg_force)

        # visualization and save image.
        explanation_mask = sp_weights_to_image_explanation(
            imgs[0], lime_weights, interpret_class[0], self.lime_base.segments
        )
        explanation_vis = overlay_threshold(imgs[0], explanation_mask)
        if visual:
            show_vis_explanation(explanation_vis)
        if save_path is not None:
            save_image(save_path, explanation_vis)

        self.lime_results['probability'] = probability
        self.lime_results['input'] = imgs[0]
        self.lime_results['segmentation'] = self.lime_base.segments
        self.lime_results['r2_scores'] = r2_scores
        self.lime_results['lime_weights'] = lime_weights

        return lime_weights




import numpy as np
import paddle
import re

from .abc_interpreter import Interpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image


class RolloutInterpreter(Interpreter):
    """
    Rollout Interpreter.

    More details regarding the Rollout method can be found in the original paper:
    https://arxiv.org/abs/2005.00928
    """

    def __init__(self,
                 paddle_model,
                 use_cuda=None,
                 device='gpu:0') -> None:
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        Interpreter.__init__(self, paddle_model, device, use_cuda)
        self.paddle_prepared = False

        # init for usages during the interpretation.
        self.block_attns = []

    def interpret(self,
                  inputs,
                  start_layer=0,
                  resize_to=224, 
                  crop_to=None,
                  visual=True,
                  save_path=None):
        """
        Main function of the interpreter.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. 
                If None, the most likely label for each image will be used. Default: None
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        Returns:
            [numpy.ndarray]: interpretations/heatmap for images
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)
        bsz = len(data)  # batch size
        save_path = preprocess_save_path(save_path, bsz)

        if not self.paddle_prepared:
            self._paddle_prepare()

        blk_attns, preds = self.predict_fn(data)

        assert start_layer < len(blk_attns), "start_layer should be in the range of [0, num_block-1]"

        label = np.array(preds).reshape((bsz,))

        all_layer_attentions = []
        for attn_heads in blk_attns:
            avg_heads = (attn_heads.sum(axis=1) / attn_heads.shape[1]).detach()
            all_layer_attentions.append(avg_heads)

        # compute rollout between attention layers
        # adding residual consideration- code adapted from https://github.com/samiraabnar/attention_flow
        num_tokens = all_layer_attentions[0].shape[1]
        batch_size = all_layer_attentions[0].shape[0]
        eye = paddle.eye(num_tokens).expand((batch_size, num_tokens, num_tokens))
        all_layer_attentions = [all_layer_attentions[i] + eye for i in range(len(all_layer_attentions))]
        matrices_aug = [all_layer_attentions[i] / all_layer_attentions[i].sum(axis=-1, keepdim=True)
                            for i in range(len(all_layer_attentions))]
        joint_attention = matrices_aug[start_layer]
        for i in range(start_layer+1, len(matrices_aug)):
            joint_attention = matrices_aug[i].bmm(joint_attention)

        # rollout
        rollout_explanation = joint_attention[:, 0, 1:].reshape((-1, 14, 14)).cpu().detach().numpy()

        # visualization and save image.
        for i in range(bsz):
            vis_explanation = explanation_to_vis(imgs[i], rollout_explanation[i], style='overlay_heatmap')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        return rollout_explanation

    def _paddle_prepare(self, predict_fn=None):
        if predict_fn is None:
            paddle.set_device(self.device)
            # to get gradients, the ``train`` mode must be set.
            # we cannot set v.training = False for the same reason.
            self.paddle_model.train()

            # def hook for attention layer
            def hook(layer, input, output):
                self.block_attns.append(output)

            for n, v in self.paddle_model.named_sublayers():
                if re.match('^blocks.*.attn.attn_drop$', n):
                    # print(n)
                    v.register_forward_post_hook(hook)

                if "batchnorm" in v.__class__.__name__.lower():
                    v._use_global_stats = True
                if "dropout" in v.__class__.__name__.lower():
                    v.p = 0

                # Report issues or pull requests if more layers need to be added.

            def predict_fn(data):
                data = paddle.to_tensor(data)
                data.stop_gradient = False
                self.block_attns = []
                out = self.paddle_model(data)
                out = paddle.nn.functional.softmax(out, axis=1)
                preds = paddle.argmax(out, axis=1)
                label = preds.numpy()

                return self.block_attns, label

        self.predict_fn = predict_fn
        self.paddle_prepared = True



import numpy as np

from tqdm import tqdm
from .abc_interpreter import InputGradientInterpreter
from ..data_processor.readers import images_transform_pipeline, preprocess_save_path
from ..data_processor.visualizer import explanation_to_vis, show_vis_explanation, save_image


class SmoothGradInterpreterV2(InputGradientInterpreter):
    """

    Smooth Gradients method solves the problem of meaningless local variations in partial derivatives
    by adding random noise to the inputs multiple times and take the average of the
    gradients.

    More details regarding the Smooth Gradients method can be found in the original paper:
    http://arxiv.org/pdf/1706.03825.pdf
    """

    def __init__(self,
                 paddle_model,
                 use_cuda=None,
                 device='gpu:0'):
        """

        Args:
            paddle_model (callable): A model with ``forward`` and possibly ``backward`` functions.
            device (str): The device used for running `paddle_model`, options: ``cpu``, ``gpu:0``, ``gpu:1`` etc.
            use_cuda (bool):  Would be deprecated soon. Use ``device`` directly.
        """
        
        InputGradientInterpreter.__init__(self, paddle_model, device, use_cuda)

    def interpret(self,
                  inputs,
                  labels=None,
                  noise_amount=0.1,
                  n_samples=50,
                  split=2,
                  resize_to=224, 
                  crop_to=None,
                  visual=True,
                  save_path=None):
        """
        Main function of the interpreter. This passes the unit tests of test_cv and test_cv_class.

        Args:
            inputs (str or list of strs or numpy.ndarray): The input image filepath or a list of filepaths or numpy array of read images.
            labels (list or tuple or numpy.ndarray, optional): The target labels to analyze. The number of labels should be equal to the number of images. If None, the most likely label for each image will be used. Default: None
            noise_amount (float, optional): Noise level of added noise to the image.
                                            The std of Guassian random noise is noise_amount * (x_max - x_min). Default: 0.1
            n_samples (int, optional): The number of new images generated by adding noise. Default: 50
            resize_to (int, optional): [description]. Images will be rescaled with the shorter edge being `resize_to`. Defaults to 224.
            crop_to ([type], optional): [description]. After resize, images will be center cropped to a square image with the size `crop_to`. 
                If None, no crop will be performed. Defaults to None.
            visual (bool, optional): Whether or not to visualize the processed image. Default: True
            save_path (str or list of strs or None, optional): The filepath(s) to save the processed image(s). If None, the image will not be saved. Default: None

        :return: interpretations/gradients for each image
        :rtype: numpy.ndarray
        """

        imgs, data = images_transform_pipeline(inputs, resize_to, crop_to)
        # print(imgs.shape, data.shape, imgs.dtype, data.dtype)  # (1, 224, 224, 3) (1, 3, 224, 224) uint8 float32

        assert len(data) == 1, "interpret each sample individually, it is optimized."

        self._build_predict_fn(gradient_of='probability')

        # obtain the labels (and initialization).
        if labels is None:
            _, preds = self.predict_fn(data, None)
            labels = preds

        labels = np.array(labels).reshape((1, ))

        # SmoothGrad
        max_axis = tuple(np.arange(1, data.ndim))
        stds = noise_amount * (
            np.max(data, axis=max_axis) - np.min(data, axis=max_axis))

        data_noised = []
        for i in range(n_samples):
            noise = np.concatenate([
                np.float32(
                    np.random.normal(0.0, stds[j], (1, ) + tuple(d.shape)))
                for j, d in enumerate(data)
            ])
            data_noised.append(data + noise)
            
        data_noised = np.concatenate(data_noised, axis=0)
        # print(data_i.shape, labels.shape)
        # print(data_noised.shape)  # n_samples, 3, 224, 224

        # splits, to avoid large GPU memory usage.
        if split > 1:
            chunk = n_samples // split
            gradient_chunks = []
            for i in range(split-1):
                gradients_i, _ = self.predict_fn(data_noised[i*chunk: (i+1) * chunk], np.repeat(labels, chunk))
                gradient_chunks.append(gradients_i)
            gradients_s, _ = self.predict_fn(data_noised[chunk*(split-1):], np.repeat(labels, n_samples - chunk*(split-1)))
            gradient_chunks.append(gradients_s)
            gradients = np.concatenate(gradient_chunks, axis=0)
        else:
            # one split.
            gradients, _ = self.predict_fn(data_noised, np.repeat(labels, n_samples))

        avg_gradients = np.mean(gradients, axis=0, keepdims=True)
        # visualize and save image.
        if save_path is None and not visual:
            # no need to visualize or save explanation results.
            pass
        else:
            save_path = preprocess_save_path(save_path, 1)
            # print(imgs[i].shape, avg_gradients[i].shape)
            vis_explanation = explanation_to_vis(imgs[i], np.abs(avg_gradients[0]).sum(0), style='overlay_grayscale')
            if visual:
                show_vis_explanation(vis_explanation)
            if save_path[i] is not None:
                save_image(save_path[i], vis_explanation)

        # intermediate results, for possible further usages.
        self.labels = labels

        return avg_gradients



import os
import os.path as osp
import numpy as np
import tqdm
import paddle
from sklearn.linear_model import Ridge
from paddle.vision.transforms import functional as F

from ..data_processor.readers import read_image, load_pickle_file
from ..common.paddle_utils import FeatureExtractor, extract_superpixel_features, get_pre_models


def data_labels(file_path_list, predict_fn, batch_size):
    print(
        "Initialization for fast NormLIME: Computing each sample in the test list."
    )

    _, h_pre_models_kmeans = get_pre_models()
    kmeans_model = load_pickle_file(h_pre_models_kmeans)
    num_features = len(kmeans_model.cluster_centers_)
    y_labels = []

    fextractor = FeatureExtractor()
    x_data = []
    tmp_imgs = []

    for each_data_ in tqdm.tqdm(file_path_list):
        image_show = read_image(each_data_)  # todo: whether add resize_to, crop_to?
        tmp_imgs.append(image_show)

        paddle.enable_static()
        feature = fextractor.forward(image_show).transpose((1, 2, 0))
        paddle.disable_static()

        img_size = (image_show.shape[1], image_show.shape[2])
        feature = F.resize(feature, img_size)

        segments = np.zeros(img_size, np.int32)
        num_blocks = 10
        height_per_i = segments.shape[0] // num_blocks + 1
        width_per_i = segments.shape[1] // num_blocks + 1

        for i in range(segments.shape[0]):
            for j in range(segments.shape[1]):
                segments[i,
                         j] = i // height_per_i * num_blocks + j // width_per_i

        X = extract_superpixel_features(feature, segments)

        try:
            cluster_labels = kmeans_model.predict(X)
        except AttributeError:
            from sklearn.metrics import pairwise_distances_argmin_min
            cluster_labels, _ = pairwise_distances_argmin_min(
                X, kmeans_model.cluster_centers_)
        x_data_i = np.zeros((num_features))
        for c in cluster_labels:
            x_data_i[c] = 1

        x_data.append(x_data_i)

        if len(tmp_imgs) == batch_size:
            outputs = predict_fn(np.concatenate(tmp_imgs, axis=0))
            y_labels.extend(outputs)
            tmp_imgs = []

    if len(tmp_imgs) > 0:
        outputs = predict_fn(np.concatenate(tmp_imgs, axis=0))
        y_labels.extend(outputs)

    return x_data, y_labels


def ridge_regressor(x_data, y_labels, softmax):
    clf = Ridge()
    clf.fit(x_data, y_labels)
    num_classes = clf.coef_.shape[0]
    global_weights_all_labels = {}

    if len(y_labels) / num_classes < 3:
        print("Warning: The test samples in the dataset is limited.\n "
              "NormLIME may have no effect on the results.\n "
              "Try to add more test samples, or see the results of LIME.")

    # clf.coef_ has shape of [len(np.unique(y_labels)), num_features]
    #
    for class_index in range(num_classes):
        w = clf.coef_[class_index]

        if softmax:
            w = w - np.max(w)
            exp_w = np.exp(w * 10)
            w = exp_w / np.sum(exp_w)

        global_weights_all_labels[class_index] = {
            i: wi
            for i, wi in enumerate(w)
        }

    return global_weights_all_labels


def precompute_global_prior(test_set_file_list,
                            predict_fn,
                            batch_size,
                            gp_method="ridge",
                            softmax=False):

    x_data, y_labels = data_labels(test_set_file_list, predict_fn, batch_size)

    if gp_method.lower() == 'ridge':
        global_weights_all_labels = ridge_regressor(x_data, y_labels, softmax)
    else:
        return None

    return global_weights_all_labels


def use_fast_normlime_as_prior(image_show, segments, label_index,
                               global_weights):
    assert isinstance(global_weights, dict)

    _, h_pre_models_kmeans = get_pre_models()

    try:
        kmeans_model = load_pickle_file(h_pre_models_kmeans)
    except:
        raise ValueError(
            "NormLIME needs the KMeans model, where we provided a default one in "
            "pre_models/kmeans_model.pkl.")

    fextractor = FeatureExtractor()
    paddle.enable_static()
    feature = fextractor.forward(image_show).transpose((1, 2, 0))
    paddle.disable_static()

    img_size = (image_show.shape[1], image_show.shape[2])
    feature = F.resize(feature, img_size)

    X = extract_superpixel_features(feature, segments)

    try:
        cluster_labels = kmeans_model.predict(X)
    except AttributeError:
        from sklearn.metrics import pairwise_distances_argmin_min
        cluster_labels, _ = pairwise_distances_argmin_min(
            X, kmeans_model.cluster_centers_)

    cluster_weights = global_weights.get(label_index, {})
    local_weights = [cluster_weights.get(k, 0.0) for k in cluster_labels]

    return local_weights
